//
// File:        XMLtoAST.java
// Package:     gov.llnl.babel.parsers.xml2
// Copyright:   (c) 2006 The Regents of the University of California
// Release:     $Name$
// Revision:    @(#) $Revision: 5496 $
// Date:        $Date: 2006-05-18 21:34:19 -0700 (Thu, 18 May 2006) $
// Description: Code to read in SIDL XML type information and create AST
// 

package gov.llnl.babel.parsers.xml2;

import gov.llnl.babel.ast.ASTNode;
import gov.llnl.babel.ast.Attribute;
import gov.llnl.babel.ast.AttributeList;
import gov.llnl.babel.ast.ClassType;
import gov.llnl.babel.ast.EnumItem;
import gov.llnl.babel.ast.Enumeration;
import gov.llnl.babel.ast.ImplementsList;
import gov.llnl.babel.ast.IntLiteral;
import gov.llnl.babel.ast.InterfaceType;
import gov.llnl.babel.ast.Method;
import gov.llnl.babel.ast.MethodName;
import gov.llnl.babel.ast.Name;
import gov.llnl.babel.ast.NamedType;
import gov.llnl.babel.ast.Package;
import gov.llnl.babel.ast.SIDLFile;
import gov.llnl.babel.ast.ScopedID;
import gov.llnl.babel.ast.Version;
import gov.llnl.babel.xml.ExceptionErrorHandler;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import org.apache.xerces.parsers.SAXParser;
import org.xml.sax.Attributes;
import org.xml.sax.EntityResolver;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;

/**
 * This class converts an XML file containing SIDL type information into
 * an equivalent AST node.
 */
public class XMLtoAST implements ContentHandler {
  private LinkedList   d_stack;
  private Locator d_locator;
  private TypeResolver d_resolver = new TypeResolverList();

  private class ElementProcessor {
    /** 
     * Determine the element name based on the class name.
     */
    public String getName() {
      String className = this.getClass().getName();
      className = className.substring(className.lastIndexOf('.')+1);
      if (className.startsWith("Start")) return className.substring(5);
      else return className.substring(3);
    }

    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {}
  }

  private class StartSymbol extends ElementProcessor { }

  private class EndSymbol extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
      try {
        NamedType nt = (NamedType)d_stack.removeLast();
        String comment = (String)d_stack.removeLast();
        Map metadata = (Map)d_stack.removeLast();
        String date = (String)d_stack.removeLast();
        Version version = (Version)d_stack.removeLast();
        Name name = (Name)d_stack.removeLast();
        nt.setName(name);
        nt.setVersion(version);
        // currently ignoring metadata & date
        nt.setDocComment(comment);
      }
      catch (ClassCastException cce) {
        throw new SAXException("XML parse stack is corrupted.");
      }
    }
  }

  private static String getRequired(Attributes attrs,
                                    String element,
                                    String key)
    throws SAXException
  {
    String value = attrs.getValue("", key);
    if (value == null) {
      throw new SAXException("XML element " + element + 
                             " is missing required attribute '" +
                             key + "'.");
    }
    return value;
  }

  private class StartSymbolName extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      d_stack.addLast(new Name(getRequired(attrs,"SymbolName","name")));
      d_stack.addLast(new Version(getRequired(attrs,"SymbolName","version")));
    }
  }

  private class EndSymbolName extends ElementProcessor { }

  private class StartMetadata extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      d_stack.addLast(getRequired(attrs, "Metadata", "date"));
      d_stack.addLast(new HashMap());
    }
  }
  
  private class EndMetadata extends ElementProcessor { }

  private class StartMetadataEntry extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      try {
        Map map = (Map)d_stack.getLast();
        map.put(getRequired(attrs, "MetadataEntry", "key"),
                getRequired(attrs, "MetadataEntry", "value"));
      }
      catch(ClassCastException cce) {
        throw new SAXException("XML stack is corrupt.");
      }
    }
  }
  
  private class EndMetadataEntry extends ElementProcessor { }

  private class StartComment extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
      d_stack.addLast(new StringBuffer());
    }
  }
  
  private class EndComment extends ElementProcessor { 
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      d_stack.addLast(d_stack.removeLast().toString());
    }
  }
  
  private class StartPackage extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      Package pkg = new Package();
      AttributeList al = pkg.getAttributeList();
      al.addAttribute(new Attribute("final",
                                    getRequired(attrs,"Package", "final")));
      d_stack.addLast(pkg);
    }
  }

  private class EndPackage extends ElementProcessor { }
  
  private class StartPackageSymbol extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException
    {
      try {
        Package pkg = (Package)d_stack.getLast();
        NamedType child = 
          d_resolver.lookupType(getRequired(attrs, "PackageSymbol", "name"),
                                attrs.getValue("", "version"));
        pkg.appendNamedType(child);
        child.setParent(pkg);
        child.accept(new RelativiseNamesVisitor(), null);
      }
      catch (UnknownTypeException ute) {
        throw new SAXException(ute);
      }
    }
  }

  private class EndPackageSymbol extends ElementProcessor { }

  private class StartEnumeration extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      d_stack.addLast(new Enumeration());
    }
  }

  private class EndEnumeration extends ElementProcessor { }

  private class StartEnumerator extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      try {
        EnumItem ei = new EnumItem();
        final String value = getRequired(attrs, "Enumerator", "value");
        ei.setName(new Name(getRequired(attrs,"Enumerator", "name")));
        if ("true".equals(getRequired(attrs, "Enumerator", "fromuser"))) {
          IntLiteral il = new IntLiteral();
          il.setInt(Integer.parseInt(value));
          ei.setExplicitValue(il);
        }
        else {
          ei.setValue(Integer.parseInt(value));
        }
        Enumeration e = (Enumeration)d_stack.getLast();
        e.addEnumItem(ei);
      }
      catch( NumberFormatException nfe) {
        throw new SAXException("Value in Enumerator is not an integer", nfe);
      }
    }
  }

  private class EndEnumerator extends ElementProcessor { }

  private class StartClass extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      ClassType ct = new ClassType();
      AttributeList al = ct.getAttributeList();
      al.addAttribute(new Attribute("abstract",
                                    getRequired(attrs,"Class", "abstract")));
      d_stack.addLast(ct);
    }
  }

  private class EndClass extends ElementProcessor { }

  private class StartExtends extends ElementProcessor { }

  private class EndExtends extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      ASTNode top = (ASTNode)d_stack.getLast();
      if (!(top instanceof ClassType)) {
        d_stack.removeLast();
        String extendName = (String)d_stack.removeLast();
        ClassType ct = (ClassType)d_stack.getLast();
        ct.setExtends(new ScopedID(extendName)); // TODO: do something version
      }
    }
  }

  private class StartImplementsBlock extends ElementProcessor { }

  private class EndImplementsBlock extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      ImplementsList il = new ImplementsList();
      while (!(d_stack.getLast() instanceof ClassType)) {
        Version v = (Version)d_stack.removeLast();
        il.addImplements(new ScopedID((String)d_stack.removeLast()));
      }
      ((ClassType)d_stack.getLast()).setImplementsList(il);
    }
  }

  private class StartInterface extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      InterfaceType it = new InterfaceType();
      d_stack.addLast(it);
    }
  }

  private class EndInterface extends ElementProcessor { }

  private class StartExtendsBlock extends ElementProcessor {}

  private class EndExtendsBlock extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      LinkedList l = new LinkedList();
      while (!(d_stack.getLast() instanceof InterfaceType)) {
        Version v = (Version)d_stack.removeLast();
        l.addLast((String)d_stack.removeLast());
      }
      // TODO: fixme
      //((InterfaceType)d_stack.getLast()).getExtends().addAll(l);
    }
  }

  private class StartAllParentClasses extends ElementProcessor {}

  private class EndAllParentClasses extends ElementProcessor { }

  private class StartAllParentInterfaces extends ElementProcessor {}

  private class EndAllParentInterfaces extends ElementProcessor { }

  private class StartMethodsBlock extends ElementProcessor {}

  private class EndMethodsBlock extends ElementProcessor { }

  private class StartMethod extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
      Method m = new Method();
      MethodName mn = new MethodName();
      mn.setShortName(new Name(getRequired(attrs, "Method", "shortname")));
      String ext = getRequired(attrs, "Method", "extension");
      if (ext.length() > 0) {
        mn.setNameExtension(new Name(ext));
      }
      m.setMethodName(mn);
      // TODO: processs copy, definition and communication attributes
      d_stack.addLast(m);
    }
  }

  private class EndMethod extends ElementProcessor { }

  private class StartArgumentList extends ElementProcessor { }

  private class EndArgumentList extends ElementProcessor { }

  private class StartThrowsList extends ElementProcessor { }

  private class EndThrowsList extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException 
    {
    }
  }

  private class StartImplicitThrowsList extends ElementProcessor { }

  private class EndImplicitThrowsList extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartAssertionList extends ElementProcessor { }

  private class EndAssertionList extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartArgument extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndArgument extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartType extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndType extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartIndex extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndIndex extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartArray extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndArray extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartAssertion extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndAssertion extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartAssertionExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndAssertionExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartBinaryExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndBinaryExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartComplexNumber extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndComplexNumber extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartMethodCall extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndMethodCall extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartSimpleExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndSimpleExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class StartUnaryExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private class EndUnaryExpression extends ElementProcessor {
    public void processElement(String namespace, String localName,
                          String qName, Attributes attrs)
      throws SAXException {
    }
  }

  private HashMap d_startProcessors;
  private HashMap d_endProcessors;

  public XMLtoAST()
  {
    ElementProcessor [] elementStarts = { 
      new StartAllParentClasses(),
      new StartAllParentInterfaces(),
      new StartArgumentList(),
      new StartArgument(),
      new StartArray(),
      new StartAssertion(),
      new StartAssertionExpression(),
      new StartAssertionList(),
      new StartBinaryExpression(),
      new StartClass(),
      new StartComment(),
      new StartComplexNumber(),
      new StartEnumeration(),
      new StartEnumerator(),
      new StartExtendsBlock(),
      new StartExtends(),
      new StartImplementsBlock(),
      new StartImplicitThrowsList(),
      new StartIndex(),
      new StartInterface(),
      new StartMetadataEntry(),
      new StartMetadata(),
      new StartMethodCall(),
      new StartMethod(),
      new StartMethodsBlock(),
      new StartPackage(),
      new StartPackageSymbol(),
      new StartSimpleExpression(),
      new StartSymbolName(),
      new StartSymbol(),
      new StartThrowsList(),
      new StartType(),
      new StartUnaryExpression()
    };
    ElementProcessor [] elementEnds = { 
      new EndAllParentClasses(),
      new EndAllParentInterfaces(),
      new EndArgumentList(),
      new EndArgument(),
      new EndArray(),
      new EndAssertion(),
      new EndAssertionExpression(),
      new EndAssertionList(),
      new EndBinaryExpression(),
      new EndClass(),
      new EndComment(),
      new EndComplexNumber(),
      new EndEnumeration(),
      new EndEnumerator(),
      new EndExtendsBlock(),
      new EndExtends(),
      new EndImplementsBlock(),
      new EndImplicitThrowsList(),
      new EndIndex(),
      new EndInterface(),
      new EndMetadataEntry(),
      new EndMetadata(),
      new EndMethodCall(),
      new EndMethod(),
      new EndMethodsBlock(),
      new EndPackage(),
      new EndPackageSymbol(),
      new EndSimpleExpression(),
      new EndSymbolName(),
      new EndSymbol(),
      new EndThrowsList(),
      new EndType(),
      new EndUnaryExpression()
    };
    int i;
    for(i = 0 ; i < elementStarts.length; ++i) {
      d_startProcessors.put(elementStarts[i].getName(), elementStarts[i]);
    }
    for(i = 0; i < elementEnds.length; ++i) {
      d_endProcessors.put(elementEnds[i].getName(), elementEnds[i]);
    }
  }


  /**
   * Convert the document to an AST.
   */
  public SIDLFile convertToAST(InputSource is, EntityResolver er) 
    throws SAXException, IOException
  {
    SAXParser parser = new SAXParser();
    d_locator = null;
    parser.setFeature("http://xml.org/sax/features/validation", true);
    parser.setErrorHandler(new ExceptionErrorHandler());
    if (er != null) {
      parser.setEntityResolver(er);
    }
    parser.setContentHandler(this);
    parser.parse(is);
    SIDLFile result = (SIDLFile)d_stack.removeLast();
    d_stack = null;
    d_locator = null;
    return result;
  }

  public void setTypeResolver(TypeResolver tr) {
    d_resolver = tr;
  }

  public void characters(char[] ch, int start, int length)
    throws SAXException
  {
  }

  public void ignorableWhitespace(char [] ch, int start, int length)
    throws SAXException
  {
    /* ignore ignorable whitespace */
  }

  public void processingInstruction(String target, String data)
    throws SAXException 
  {
    /* ignore all processing instructions */
  }

  public void setDocumentLocator(Locator locator)
  {
    d_locator = locator;
  }

  public void skippedEntity(String name)
    throws SAXException
  {
    throw new SAXException("Skipping entity '" + name + "'.");
  }

  public void startDocument()
  {
    d_stack = new LinkedList();
  }

  public void endDocument()
  {
  }

  public void startElement(String namespaceURI, String localName, String qName,
                    Attributes attrs)
    throws SAXException
  {
    ElementProcessor ep = (ElementProcessor)d_startProcessors.get(localName);
    if (ep != null) {
      ep.processElement(namespaceURI, localName, qName, attrs);
    }
    else {
      throw new SAXException("Unknown XML element " + localName + 
                             " while parsing type.");
    }
  }

  public void endElement(String namespaceURI, String localName, String qName)
    throws SAXException
  {
    ElementProcessor ep = (ElementProcessor)d_endProcessors.get(localName);
    if (ep != null) {
      ep.processElement(namespaceURI, localName, qName, null);
    }
    else {
      throw new SAXException("Unknown XML element " + localName + 
                             " while parsing type.");
    }
  }

  /**
   * SIDL XML isn't expected to have prefix mappings.
   */
  public void startPrefixMapping(String prefix, String uri)
    throws SAXException
  {
    throw new SAXException("Unexpected start of prefix-URI namespace mapping.");
  }


  /**
   * SIDL XML isn't expected to have prefix mappings.
   */
  public void endPrefixMapping(String prefix)
    throws SAXException
  {
    throw new SAXException("Unexpected end of prefix-URI namespace mapping.");
  }
  
}
